# ---------------------------------------------------------------------------------------------------------------
# piwiPre project
# This program and library is licenced under the European Union Public Licence v1.2 (see LICENCE)
# developed by fabien.battini(at)gmail.com
# ---------------------------------------------------------------------------------------------------------------
import sys
# pip install GitPython
import git
import re
import os
import argparse
import datetime

# this file is run by tox.
# All installation commands are executed using {toxinidir} (the directory where tox.ini resides) as CWD # noqa
# to test without tox, run: 'cd .. ; python piwiPre/pwpPatcher.py  '                                    # noqa
# the following line is mandatory to execute from shell
# tox does it automatically, so not necessary if running only from tox.
sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from piwiPre.pwpParser import build_official_rst


class PwpPatcher:
    def __init__(self, arguments=None, standalone=True):
        if not standalone:
            return
        self.parser = argparse.ArgumentParser(description='manages tags and modifies accordingly files that want it ')
        self.parser.add_argument('--Major', '-M', help="Promotes as Major release i.e. X++.",
                                 action='store_true')
        self.parser.add_argument('--minor', '-m', help="Promotes as minor version ie x.Y++",
                                 action='store_true')
        self.parser.add_argument('--patch', '-p', help="Increments patch version ie x.y.Z++",
                                 action='store_true')
        self.parser.add_argument('--commit', '-c', help="Commits, implicit if -M, -m, -p",
                                 action='store_true')
        self.parser.add_argument('--full-help', help="more explanations",
                                 action='store_true')
        self.parser.add_argument('--print', help="prints the last tag spec and exits",
                                 action='store_true')
        self.parser.add_argument('--download', '-d', help="patches download.rst",
                                 action='store_true')
        self.parser.add_argument('--autotest', help="autotest, no real action",
                                 action='store_true')

        self.repo = git.Repo('.')
        tags = sorted(self.repo.tags, key=lambda t: t.commit.committed_datetime)
        if len(tags) > 0:
            latest_tag = tags[-1]
            tag_version = str(latest_tag)
        else:
            latest_tag = None
            tag_version = "0.0.0"

        full_tag = re.sub(r"[^\d.]*", '', tag_version)
        m = re.match(r"(\d+)\.*(\d*)\.*(\d*)", full_tag)
        self.release = int(m.group(1)) if m.group(1) else 0
        self.version = int(m.group(2)) if m.group(2) else 0
        self.patch = int(m.group(3)) if m.group(3) else 0
        self.spec = full_tag
        if latest_tag:
            self.date = latest_tag.commit.committed_datetime.strftime("%m/%d/%Y %H:%M:%S")
        else:
            self.date = datetime.datetime.now()
        self.update_spec()
        self.args = self.parser.parse_args() if arguments is None else self.parser.parse_args(arguments)

    def update_spec(self):
        self.spec = str(self.release)
        if self.version != 0 or self.patch != 0:
            self.spec += '.' + str(self.version)
            if self.patch != 0:
                self.spec += '.' + str(self.patch)

    def manage_tags(self):

        if self.args.patch:
            self.patch += 1

        if self.args.minor:
            self.version += 1
            self.patch = 0

        if self.args.Major:
            self.release += 1
            self.version = 0
            self.patch = 0

        if self.args.Major or self.args.minor or self.args.patch:
            n = datetime.datetime.now()
            self.date = n.strftime("%m/%d/%Y %H:%M:%S")
            self.update_spec()

            print(f"msg     new tag '{self.spec}'")
            return True
        return False

    @staticmethod
    def open(file_name, mode):
        try:
            ins = open(file_name, mode, encoding="utf-8")
        except OSError as error:
            print(f"FATAL ERROR: Cannot open('{file_name}',{mode}) : {str(error)}, aborting")
            exit(-1)
        return ins

    @staticmethod
    def rename(old, new):
        try:
            os.rename(old, new)
        except OSError as error:
            print(f"FATAL ERROR: Cannot rename('{old}','{new}') : {str(error)}, aborting")
            exit(-1)
        return

    def update_file(self, filename: str, dico: dict, duplicate=False, autotest=False):

        if autotest:
            new_file = f"tests/results/{os.path.basename(filename)}.autotest"
        else:
            new_file = filename+'.new'
        old_file = filename+'.bak'

        ins = self.open(filename, "r")
        content = ins.readlines()
        ins.close()

        outs = self.open(new_file, "w")
        useful = False
        previous = []
        for line in content:
            modified = line
            for before, after in dico.items():
                modified = re.sub(before, after, modified)
            if modified not in previous:
                previous.append(modified)
            else:
                # we have already done that modification once, we do not do it twice
                modified = line
            outs.writelines(modified)

            if modified != line:
                useful = True
                if duplicate:
                    outs.writelines(line)

        outs.close()

        if autotest:
            print(f"msg     autotest file '{new_file}' generated")
            return

        if not useful:
            os.remove(new_file)
            print(f"msg     file '{filename}' does not need patch")
            return

        if os.path.isfile(old_file):
            os.remove(old_file)
        self.rename(filename, old_file)
        self.rename(new_file, filename)
        print(f"msg     file '{filename}' patched")

    # @staticmethod
    # def update_from_py(filename, source):
    #     if os.path.getmtime(filename) > os.path.getmtime(source):
    #         print(f"msg     file '{filename}' is older than source '{source}': patch useless")
    #         return
    #     args_ini_main(['--build-rst-file', '--rst-file', filename])

    def update_from_source(self, filename, source, autotest: bool):

        if not autotest and os.path.getmtime(filename) > os.path.getmtime(source):
            print(f"msg     file '{filename}' is older than source '{source}': patch useless")
            return

        with self.open(source, "r") as ins:
            content = ins.readlines()

        if autotest:
            new_file = f"tests/results/{os.path.basename(filename)}.autotest"
        else:
            new_file = filename + '.new'
        old_file = filename + '.bak'

        outs = self.open(new_file, "w")
        do_add = True
        with self.open(filename, "r") as ins:
            for line in ins:
                if do_add:
                    outs.write(line)

                    if re.match(r' *def print\(.*\):\n', line):
                        for cl in content:
                            patched = re.sub('"', '\\"', cl[:-1])
                            outs.write(f'        print("{patched}")\n') # add \t\t\t # noqa ???
                        outs.write('    # End of patched text\n')
                        do_add = False
                else:
                    if re.match(r' +# End of patched text\n', line):
                        do_add = True
        outs.close()

        if autotest:
            print(f"msg     autotest file '{new_file}' generated")
            return

        if os.path.isfile(old_file):
            os.remove(old_file)
        self.rename(filename, old_file)
        self.rename(new_file, filename)
        print(f"msg     file '{filename}' patched")

    def run(self):
        if self.args.autotest:
            self.args.full_help = True

        if self.args.print:
            print(self.spec)
            return

        print(f"msg     initial tag '{self.spec}' at '{self.date}'")

        if self.args.full_help:
            print("if -M, -m or -p, updates the current tag")
            print("verifies that the current tag is writen in piwiPre/pwpVersion.py")
            print("verifies that piwiPre/pwpLicence.py is up to date vs source")
            print("if ")
            print("   - the new tag is a Major or a minor (but *not* a patch), then the corresponding exe and module")
            print("     will be published, assuming here that they have been tested OK)")
            print("     therefore download.rst is also modified")
            print("   - or -d")
            print("then download.rst is updated")

            print("if a new tag or -c, commits ALL modified files")
            print("if a new tag puts the tag LOCALLY on git, but *not* remotely on GitLab")

        new_tag = self.manage_tags()
        if self.args.autotest:
            self.args.Major = True
            self.manage_tags()
            self.args.Major = False
            self.args.minor = True
            self.manage_tags()
            self.args.minor = False
            self.args.patch = True
            new_tag = self.manage_tags()

        self.update_from_source('piwiPre/pwpLicence.py', 'LICENCE', self.args.autotest)

        # self.update_from_source('piwiPre/pwpIniFile.py', 'source/usage/configuration.rst', config=True)

        self.update_file('piwiPre/pwpVersion.py',
                         {r"help = '([^']*)'": f"help = '{self.spec} at {self.date}'",
                          r"spec = '([^']*)'": f"spec = '{self.spec}'",
                          r"release = (\d*)": f"release = {self.release}",
                          r"version = (\d*)": f"version = {self.version}",
                          r"patch = (\d*)": f"patch = {self.patch}",
                          r"date = '([^']*)'": f"date = '{self.date}'",
                          }, autotest=self.args.autotest)

        # - Windows one-file exe on gitlab artifacts:
        # `piwiPre-1.8.3.exe
        # <https://gitlab.com/api/v4/projects/22464405/packages/generic/pyxlSql/1.8.3/pyxlSql-1.8.3.exe>`_

        prologue = '- Windows one-file exe on gitlab artifacts:'
        epilogue = '<https://gitlab.com/api/v4/projects/22464405/packages/generic/piwiPre'

        if self.args.download or (new_tag and self.patch == 0) or self.args.autotest:
            self.update_file('source/download.rst',
                             {
                                 r".*Windows one-file exe.*": f"{prologue} `piwiPre-{self.spec}.exe " +
                                                              f"{epilogue}/{self.spec}/piwiPre-{self.spec}.exe>`_"
                             },
                             duplicate=True, autotest=self.args.autotest)

        self.update_file('version.txt', {r"(.+)":  self.spec}, autotest=self.args.autotest)
        build_official_rst(self.args.autotest)

        if self.args.autotest:
            print(f"msg     final tag '{self.spec}' at '{self.date}'")
            return

        if new_tag or self.args.commit:
            commit = self.repo.git.commit('-a', '-m', f'set tag {self.spec}', '--date', self.date)
            print(f"msg     DONE: git commit '{str(commit)}' ")

        if new_tag:
            self.repo.git.tag(self.spec)
            print(f"msg     DONE: git tag '{self.spec}'")


def patcher_main(arguments):
    arguments = arguments or []
    print('msg     --------------- starting patcher_main')
    my_patcher = PwpPatcher(arguments=arguments)
    my_patcher.run()
    print('msg     --------------- ending patcher_main')


if __name__ == "__main__":
    sys.exit(patcher_main(sys.argv[1:]))



# ---------------------------------------------------------------------------------------------------------------
# piwiPre project
# This program and library is licenced under the European Union Public Licence v1.2 (see LICENCE)
# developed by fabien.battini(at)gmail.com
# ---------------------------------------------------------------------------------------------------------------

import argparse
import re
import sys
import os
import pprint
import datetime

from piwiPre.pwpConfig import PwpConfig
from piwiPre.pwpErrors import PwpConfigError, LOGGER


class ArgHeader:
    def __init__(self, config: str):
        self.config = config

    @staticmethod
    def is_in_config():
        return False

    def write_rst(self, stream):
        stream.write("\n")
        for line in self.config.splitlines():
            stream.write(line + '\n')
        stream.write("\n")

    def write_ini_file(self, stream):
        stream.write("\n")
        for line in self.config.splitlines():
            stream.write('# ' + line + '\n')
        stream.write("\n")


class ArgIniItem:
    def __init__(self, name: str, config: str, location: str, help_str: str, default, arg, i_type):
        self.name = name
        self.config = config
        self.location = location
        self.help = help_str
        self.default = default
        self.arg = arg
        self.i_type = i_type
        if self.i_type == dict and self.default == '':
            self.default = {}
        if not isinstance(self.default, self.i_type) and self.default is not None:
            raise PwpConfigError(f"Parameter '{name}' default value {default} should have type {str(i_type)} ")

    def is_in_config(self):
        return self.location != 'args'

    def write_rst(self, stream):

        default = self.get_ini_value(self.default, "   ", ini_file=False)
        # default = 'false' if default is False else 'true' if default is True else default

        location = 'configration files only' if self.location == 'config' else \
            'cmd-line arguments only' if self.location == 'args' else 'both configuration files and cmd-line arguments'
        stream.write(f"\n**{self.name}** : {default}\n\n")
        stream.write(f"   where: {location}\n\n")
        stream.write(f"   {self.help}\n")
        for line in self.config.splitlines():
            stream.write(f"   {line}\n")
        stream.write("\n")

    def get_ini_value(self, item, prefix, level=0, ini_file=True):
        if item is False:
            return 'false'
        if item is True:
            return 'true'
        if isinstance(item, str):
            if re.search(r'[+\-/]', item):
                return "'" + item + "'"
            return item

        if type(item) is dict:
            if len(item.keys()) == 0:
                return ""
            if level == 0:
                if ini_file:
                    res = '\n'
                else:
                    res = '\n ::\n\n'
            else:
                res = '\n'
            for key, value in item.items():
                k = self.get_ini_value(key, "", level=level+1, ini_file=ini_file)
                val = self.get_ini_value(value, prefix + "   ", level=level+1, ini_file=ini_file)
                res += f"{prefix}   {k} : {val}\n"
            if level == 0:
                if not ini_file:
                    res += "\n\n#  *(end of the structure)*\n"
            return res

        return str(item)

    def write_ini_file(self, stream):
        default = self.get_ini_value(self.default, "   ", ini_file=True)

        location = 'configration files only' if self.location == 'config' else \
            'cmd-line arguments only' if self.location == 'args' else 'both configuration files and cmd-line arguments'
        if self.location == 'args':
            header = '# '
            name = '--' + self.name
        else:
            header = ''
            name = self.name

        stream.write(f"\n{header}{name} : {default}\n")
        stream.write(f"#   where: {location}\n\n")
        stream.write(f"#   {self.help}\n")
        for line in self.config.splitlines():
            stream.write(f"#   {line}\n")
        stream.write("\n")


class PwpArgsIni(argparse.ArgumentParser):

    def __init__(self, **_kwargs):
        super().__init__(prog='piwiPre', allow_abbrev=False, exit_on_error=False)
        self.args_dico = {}
        # self.args = None # the actual args passed on cmd line
        self.items_list = []
        self.home_config = None

    def add_header(self, prologue: str):
        item = ArgHeader(prologue)
        self.items_list.append(item)

    def add_item(self, name_or_flags: str, **kwargs):
        # location should be 'config' or 'args'. any other value = 'both'. default = 'both'
        location = 'both'
        if 'location' in kwargs:
            location = kwargs['location']
            kwargs.pop('location')
        config = ''
        if 'config' in kwargs:
            config = kwargs['config']
            kwargs.pop('config')
        help_str = '' if 'help' not in kwargs else kwargs['help']
        default = '' if 'default' not in kwargs else kwargs['default']
        if 'pwp_type' in kwargs:
            i_type = kwargs['pwp_type']
            kwargs.pop('pwp_type')
        else:
            i_type = str

        arg = super().add_argument('--'+name_or_flags, **kwargs) if location != 'config' else None
        item = ArgIniItem(name_or_flags, config, location, help_str, default, arg, i_type)
        self.args_dico[name_or_flags] = item
        self.items_list.append(item)

    def build_rst(self, filename: str):
        abs_path = os.path.abspath(filename)
        LOGGER.debug(f"Build RST file {abs_path}")
        with open(filename, 'w', encoding="utf-8") as f:
            start = datetime.datetime.now()
            f.write(f".. comment : CAVEAT: This text is automatically generated by pwpPatcher.py on {start}\n")
            f.write(".. comment :         from the code in pwpParser.py\n")
            for item in self.items_list:
                item.write_rst(f)

    def build_ini_file(self, filename: str):
        with open(filename, 'w', encoding="utf-8") as f:
            for item in self.items_list:
                item.write_ini_file(f)

    def build_initial_config(self):
        dico = {}
        for v in self.items_list:
            if v.is_in_config():
                dico[v.name] = v.default
        dico['help'] = None
        res = PwpConfig(content_str=None, filename="INITIAL VALUES", dico=dico)
        return res

    def parse_args_and_ini(self, program: str, ini_to_parse, arguments, with_config=True):
        initial_config = None
        real_args = arguments if arguments is not None else []

        # check if actual arguments prevent from using the default .ini
        skip_next = False
        for i in range(len(real_args)):
            if skip_next:
                skip_next = False

            elif real_args[i] == '--ini-file':
                if i >= len(real_args):
                    raise PwpConfigError("ERROR: --ini-file without a value")
                new_ini_to_parse = real_args[i + 1]
                ini_to_parse = os.path.basename(new_ini_to_parse)
                skip_next = True

            elif real_args[i] == '--chdir':
                if i >= len(real_args):
                    raise PwpConfigError("ERROR: --chdir without a value")
                new_dir = real_args[i + 1]
                if not os.path.isdir(new_dir):
                    raise PwpConfigError(f"--chdir '{new_dir}' : non existing directory")

                LOGGER.msg(f"chdir '{new_dir}'")
                os.chdir(new_dir)
                skip_next = True
            elif real_args[i] == '--help':
                self.print_usage()
                # FIXME: if we do exit(0) or raisePwpConfigError, (on program_3) on test 400, Paramiko exits on error!
                #        msg is:  'ValueError: I/O operation on closed file'
                #        this seems to be an issue with stdin.
                # Trick: We return None, so that the program is ended gracefully
                return None
            else:
                skip_next = False

        if with_config:
            initial_config = self.build_initial_config()

        home_ini_path = os.path.expanduser("~") + '/.' + ini_to_parse
        config = initial_config

        if with_config and os.path.isfile(home_ini_path):
            first_ini = PwpConfig.parse_ini_file(home_ini_path)
            # if not ACTOR.is_mode_protected(home_ini_path):
            #    raise PwpError(f"HOME ini file {home_ini_path} MUST be protected by chmod 0x600 o 0x400")
            # chmod has limited meaning in windows universe
            config = first_ini.merge_ini(initial_config)

        self.home_config = config

        if with_config and os.path.isfile(ini_to_parse):
            first_ini = PwpConfig.parse_ini_file(ini_to_parse)
            config = first_ini.merge_ini(config)

        LOGGER.msg(f"{program}: reading configuration from cmd-line arguments")
        try:
            args = super().parse_args(args=real_args)
        except argparse.ArgumentError:
            super().print_help()
            raise PwpConfigError("Error on arguments, internal")
        except SystemExit:
            super().print_help()
            # return None
            raise PwpConfigError("Error on arguments")

        if with_config:
            config = config.merge_ini_args(args, real_args)
        else:
            config = PwpConfig.args_to_dict(args)

        return config


def args_ini_main(arguments):
    parser = PwpArgsIni()
    parser.add_header('Unified parser for .ini files and cmdline flags')
    parser.add_header('===============================================')
    parser.add_header('\nParameters when executing pwpArgsIni as a program\n')
    parser.add_item('build-ini-file',
                    help='builds the ini-file argument',
                    action='store_true',
                    location='args')
    parser.add_item('ini-file',
                    help='sets the ini-file to build',
                    action='store',
                    default="test.ini")
    parser.add_item('build-rst-file',
                    help='builds the rst-file argument',
                    action='store_true',
                    location='args')
    parser.add_item('rst-file',
                    help='sets the rst-file to build',
                    action='store',
                    default="test.rst")
    parser.add_item('dump-config',
                    help='dumps the configuration and exits',
                    action='store_true',
                    location='args')
    parser.add_item('auto-test',
                    help='performs the auto-test and exits',
                    action='store_true',
                    location='args')

    # check if actual arguments ask for new configuration items just for test

    if '--auto-test' in arguments:
        # The following data is fake, just to test if args-ini works OK
        parser.add_header("""
#######################
Flags and configuration
#######################""")
        parser.add_header("""
File usage
==========

This file is the default configuration of piwiPre.

Unless stated otherwise, the  configuration items have a command line argument counterpart, 
with the same name, starting with -- .

The default value is given as an argument.

The configuration file uses the yaml syntax,
and uses pyYaml  to read/write the configuration file""")
        parser.add_item('version', help="Prints piwiPre version number and exits.",
                        action='store_true', location='args')

        parser.add_header("""
Management of directories
=========================""")

        parser.add_item('triage',
                        help='Sets the root directory for TRIAGE pictures to manage.',
                        action='store',
                        default='TRIAGE',
                        config="""
- value = 'directory': Sets the root directory for TRIAGE pictures to manage
- value = None: renaming  has already been done, so the TRIAGE directory is not processed
""")

        parser.add_item('month-name',
                        help='The name for each month, used to compute month_name.',
                        action='store',
                        pwp_type=list,
                        default=['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'],
                        location='config')

        parser.add_item('piwigo-thumbnails',
                        help='A dictionary if thumbnail specifications',
                        pwp_type=dict,
                        default={
                            "{f}-sq.jpg": {'width': 120, 'height': 120, 'crop': True},
                            "{f}-th.jpg": {'width': 144, 'height': 144, 'crop': False},
                            "{f}-me.jpg": {'width': 792, 'height': 594, 'crop': False},
                            "{f}-cu_e250.jpg": {'width': 250, 'height': 250, 'crop': True},
                        },
                        location='config')
        parser.add_item('dates',
                        help='A dictionary of dates corrections',
                        action='store',
                        pwp_type=dict,
                        default={},
                        location='config')
        parser.add_item('verify-albums',
                        help='true/false/list of directories in ALBUMS to be processed ',
                        action='append',
                        pwp_type=list,
                        default=[])
        parser.add_item('process-rename',  # is used to test ambiguous arguments
                        help='Enables files renaming',
                        action='store',
                        choices=['true', 'false'],
                        default='false')
    # end of auto-test case
    config = parser.parse_args_and_ini("autotest", "tests.ini", arguments)

    if config['auto-test']:
        rst = config['rst-file'] or "../results/test-result.rst"
        ini = config['ini-file'] or "../results/test-result.ini"
        parser.build_rst(rst)
        parser.build_ini_file(ini)
        pprint.pprint(config)
        parser.print_help()
        return
    if config['build-rst-file']:
        parser.build_rst(config['rst-file'])
    if config['build-ini-file']:
        parser.build_ini_file(config['ini-file'])
    if config['help']:
        parser.print_help()


# by default, --auto-test is launched from the tests/argsini directory  # noqa

if __name__ == "__main__":
    sys.exit(args_ini_main(sys.argv[1:]))



# ---------------------------------------------------------------------------------------------------------------
# piwiPre project
# This program and library is licenced under the European Union Public Licence v1.2 (see LICENCE)
# developed by fabien.battini(at)gmail.com
# ---------------------------------------------------------------------------------------------------------------

import re
import os
import shutil
import time

# pip install fabric
# doc: https://docs.fabfile.org/en/stable/
# doc: https://www.paramiko.org/
# doc: https://help.ubuntu.com/community/SSH/OpenSSH/Keys
import fabric


# pip install requests
# doc: https://requests.readthedocs.io/en/latest/
import requests
import requests.cookies

import urllib3.exceptions
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
# Remove useless warning on https certificates

from piwiPre.pwpErrors import LOGGER, PwpError


class PwpActor:
    allowed_chars = r"a-zA-Z0-9\-_.&@~!,;+°()àâäéèêëïîôöùûüÿçñÀÂÄÉÈÊËÏÎÔÖÙÛÜŸÇÑ "  # noqa
    tmp_dir = '.piwiPre.tmp'

    def __init__(self):
        self.dryrun = False
        # values that are cached from the 1st config used
        self.print_debug = False
        self.piwigo_user = None
        self.piwigo_pwd = None
        self.piwigo_host = None
        self.piwigo_port = None

        # management of ssh/sftp
        self.remote_user = None
        self.remote_pwd = None
        self.remote_host = None
        self.remote_port = None
        self.ls_command = None
        self.ls_output = None
        self.connection = None
        self.session = None
        self.remote_uname = None

        # end of cache
        self.piwigo_session = None
        self.dir_made = []
        self.dir_numbers = {}

        self.remote_ls_cache = {}

    def configure(self, config):
        self.piwigo_host = config['piwigo-host']
        self.piwigo_port = int(config['piwigo-port'])
        self.piwigo_user = config['piwigo-user']
        self.piwigo_pwd = config['piwigo-pwd']
        self.print_debug = config['debug']
        self.dryrun = config['dryrun']

        self.connect_ssh(config)

    def reset_data(self):
        self.dryrun = False
        # values that are cached from the 1st config used
        self.print_debug = False
        self.piwigo_user = None
        self.piwigo_pwd = None
        self.piwigo_host = None
        self.piwigo_port = None
        # end of cache
        self.piwigo_session = None
        self.dir_made = []
        self.dir_numbers = {}

        self.connection = None
        self.session = None
        self.remote_ls_cache = {}

    def mkdirs(self, dir_name, forced=False):
        dir_name = dir_name.rstrip('/')
        dir_name = dir_name or '.'
        if os.path.isdir(dir_name):
            return
        if self.dryrun and not forced:
            LOGGER.debug(f"Would makedirs '{dir_name}'")
            return

        os.makedirs(dir_name, exist_ok=True)
        if os.path.isdir(dir_name):
            LOGGER.debug(f"makedirs '{dir_name}'")
        else:
            LOGGER.debug(f"FAIL: makedirs '{dir_name}'")

    def copy(self, src, dst, forced=False):
        """
        copy src to dst, unless dryrun is True
        :param src: file to copy
        :param dst: destination
        :param forced: if True, copy is always done, if False, do not copy if dryrun is True
        :return: None
        """
        base = os.path.dirname(dst)
        self.mkdirs(base, forced)

        if self.dryrun and not forced:
            LOGGER.debug(f"Would copy '{src}' ->  '{dst}'")
            return

        shutil.copy2(src, dst)  # preserve metadata
        if os.path.isfile(dst):
            LOGGER.debug(f"copy '{src}' ->  '{dst}'")
        else:
            LOGGER.debug(f"FAIL:copy '{src}' ->  '{dst}'")

    def copytree(self, src, dst):

        if self.dryrun:
            LOGGER.debug(f"Would copytree'{src}' ->  '{dst}'")
            return

        shutil.copytree(src, dst, dirs_exist_ok=True)
        if os.path.isdir(dst):
            LOGGER.debug(f"copytree '{src}' ->  '{dst}'")
        else:
            LOGGER.debug(f"FAIL:copytree '{src}' ->  '{dst}'")

    def move(self, src, dst, forced=False):
        if self.dryrun and not forced:
            LOGGER.debug(f"Would move file '{src}' -> '{dst}'")
            return

        # src_stamp = self.timestamp(src)

        base = os.path.dirname(dst)
        self.mkdirs(base)

        if os.path.isfile(dst):
            os.remove(dst)

        shutil.move(src, dst)
        if os.path.isfile(dst):
            LOGGER.debug(f"move file '{src}' -> '{dst}'")
            # dst_stamp = self.timestamp(dst)
            # if dst_stamp != src_stamp:
            #    self.msg(f"TIMESTAMPS '{src_stamp}' -> '{dst_stamp}'")
        else:
            LOGGER.debug(f"FAIL:move file '{src}' -> '{dst}'")

    def delete(self, src, forced=False):
        if not os.path.isfile(src):
            return

        if self.dryrun and not forced:
            LOGGER.debug(f"Would delete '{src}' ")
            return

        os.remove(src)
        if os.path.isfile(src):
            LOGGER.debug(f"FAIL:  delete '{src}'")
        else:
            LOGGER.debug(f"delete file '{src}'")

    def rmtree(self, src):
        if not os.path.isdir(src):
            return

        if self.dryrun:
            LOGGER.debug(f"would remove tree '{src}' ")
            return

        shutil.rmtree(src)
        if os.path.isdir(src):
            raise PwpError(f"FAIL:  remove tree '{src}'")
        else:
            LOGGER.debug(f"remove tree  '{src}'")

    # @staticmethod
    # def is_mode_protected(path):
    #     if os.path.isfile(path):
    #         mode = stat.S_IMODE(os.stat(path).st_mode)
    #         return mode & (stat.S_IRWXO | stat.S_IRWXG) == 0
    #    return False
    # chmod has limited meaning in windows universe

    @staticmethod
    def open(filename: str, mode: str):
        # if self.dryrun:
        #     LOGGER.debug(f"open '{filename}' ")
        #     return

        if mode == 'r' and not os.path.isfile(filename):
            raise PwpError(f"ERROR reading non-existing file {filename}")
        if mode == 'rb':
            return open(filename, mode)
        return open(filename, mode, encoding="utf-8")

    @staticmethod
    def get_last_numerical_field(template: str):
        items_list = re.findall('{[^}]*}', template)
        non_numerical = ['{a', '{month_name}', '{base}', '{author}', '{suffix}', '{file}']
        while items_list:
            res = items_list.pop()
            if res not in non_numerical:
                return res[1:-1]  # skip {}
        return None

    def get_info_from_format(self, template: str, src: str):
        """
        Extract information from src according to the descriptor
        :param template: a string that describes the information format
        :param src: the source with information
        :return: a dictionary

        Assuming that template is a reasonably simple format,
        notably that the same field does not occur 2 times.

        Assuming also that the possible items within template are all known,
        which is the case in piwiPre.

        If src is the result of formatting template with some values,
        then we can find back the values, provided the string is simple enough.
        This can even been done independently of the order of the fields in the template,
        because we can find the order by analysing template.
        """
        items_list = re.findall('{[^}]*}', template)  # the list of all items to find, e.g. '{Y}'
        # here, we have all items possibly found in piwiPre
        trans = {
            'size': r"(\d+)",      # noqa
            'Y': r"(\d\d\d\d)",
            'm': r"(\d\d)",
            'd': r"(\d\d)",
            'H': r"(\d\d)",
            'M': r"(\d\d)",
            'S': r"(\d\d)",
            'count': r"(\d+)",
            'z': r"(\+?\-?\d\d\d\d)",
            'a': r"(am|pm)",
            'month_name': '([' + self.allowed_chars + ']+)',
            'base': '([' + self.allowed_chars + ']+)',
            'author': r'(\.*)',
            'suffix': r'(\.*)',
            'file': r"(.*?)",  # noqa
        }
        str_format = template.format(**trans)
        res = re.match(str_format, src)
        dico = None
        if res:
            dico = {}
            for field in trans.keys():
                ff = '{'+field+'}'
                dico[field] = res.group(items_list.index(ff) + 1) if ff in items_list else None
        return dico

    def authenticate_piwigo(self, forced=False):
        if self.piwigo_session is not None and forced is None:
            return
        LOGGER.debug(f"Authenticate to piwigo, host = '{self.piwigo_host}' user = '{self.piwigo_user}'")
        redirect = '/piwigo/admin.php?page=site_update&site=1'
        piwigo_url = "https://" + self.piwigo_host + "/piwigo/identification.php"
        post_info = {
            'username': self.piwigo_user,
            'password': self.piwigo_pwd,
            'remember_me': "1",
            'redirect': redirect,
            'login': 'To Validate',
        }
        # https://requests.readthedocs.io/en/latest/user/advanced/
        self.piwigo_session = requests.Session()
        required_args = {
            'name': 'pwg_id',
            'value': 'A fake value to start the process'
        }
        optional_args = {}
        my_cookie = requests.cookies.create_cookie(**required_args, **optional_args)
        self.piwigo_session.cookies.set_cookie(my_cookie)
        self.piwigo_session.post(piwigo_url, data=post_info, verify=False)

    def maybe_authenticate(self, req):
        if '<a href="identification.php?redirect=' in req.text:
            # piwigo asks for authentication
            self.authenticate_piwigo(forced=True)
            req = self.piwigo_session.get(req.url, verify=False)
        return req

    def get_piwigo_dirs(self):
        self.authenticate_piwigo()
        LOGGER.debug("get_piwigo_dirs")
        piwigo_url = "https://" + self.piwigo_host + "/piwigo/admin.php?page=site_update&site=1"
        req = self.piwigo_session.get(piwigo_url, verify=False)
        req = self.maybe_authenticate(req)
        seen = False
        dico = {}
        # CAVEAT: the number of &nbsp; gives the level of the directory
        # O: Root dir, typically photo
        # 3: first level
        # 6: 2nd level etc
        path = []
        level = 0
        for line in req.text.splitlines():
            if seen:
                res = re.search(r'<option value="(\d+)">((&nbsp;)*)- (.*)</option>', line)
                if res:
                    item = res.group(4)
                    new_level = int(res.group(2).count("&nbsp;")/3) + 1
                    if new_level == level+1:
                        pass
                        # entered a deeper directory
                    elif new_level > level+1:
                        LOGGER.warning("INTERNAL ERROR: jump between directories")
                    else:  # new_level <= level
                        for i in range(0, level - new_level+1):
                            path.pop()
                    d = ''
                    for f in path:
                        d += f + '/'
                    dico[d + item] = res.group(1)
                    path.append(item)
                    level = new_level

                else:
                    if re.search(r'</select>', line) is not None:
                        seen = False
            else:
                seen = re.search(r'<select class="categoryList" name="cat"', line) is not None
        return dico

    def synchronize_piwigo_dir(self, src: str, with_files: bool):
        # https://fr.piwigo.org/forum/viewtopic.php?id=16532
        LOGGER.debug(f"synchronize_piwigo_dir '{src}'")
        if self.dryrun:
            return
        if self.piwigo_host is None or self.piwigo_user is None:
            LOGGER.info(f"synchronize_piwigo_dir '{src}' impossible : no host configuration")
            return
        self.dir_numbers = self.get_piwigo_dirs()
        dirs = self.dir_numbers
        if 'photo/' + src not in dirs.keys():
            LOGGER.info(f"dir '{src}' not in piwigo managed directories")
            return
        dir_nb = dirs['photo/' + src]

        post_info = {
            'sync': 'files' if with_files else 'dirs',
            'display_info': 1,
            'privacy_level': 4,
            'sync_meta': 1,  # Value is unclear
            "cat": dir_nb,
            "subcats-included": 1,  # noqa
            "submit": "submit",
        }
        piwigo_url = "https://" + self.piwigo_host + "/piwigo/admin.php?page=site_update&site=1"
        req = self.piwigo_session.post(piwigo_url, data=post_info, verify=False)
        # TODO: parse here to get new directories and verify src is in the list
        t = req.text
        LOGGER.info(f"dir '{src}' synchronized with piwigo")
        return t

    # ----------------------------------------------------------------------
    # management of ssh/sftp

    @staticmethod
    def build_timestamp(filename: str):
        file_time = os.path.getmtime(filename)
        timestamp = time.strftime("%Y/%m/%d-%H:%M:%S", time.localtime(file_time))
        return timestamp

    @staticmethod
    def timestamp_from_ls(d: dict):
        timestamp = f"{d['Y']}/{d['m']}/{d['d']}-{d['H']}:{d['M']}:{d['S']}"
        return timestamp

    def remote_run(self, cmd: str):
        if self.dryrun:
            return None

        if self.connection:
            res = self.connection.run(cmd, hide=True, warn=True, encoding='utf8')
            if not res.ok:
                LOGGER.info(f"CAVEAT remote '{cmd}' returned {res.stderr}")
            return res
        raise PwpError("trying to run a ssh command without a ssh connection", cmd)

    @staticmethod
    def my_decode(item: str) -> str:
        """
        Manages the unknown encoding used by ls - paramiko - fabfile  
        returns the decoded string
        only chars in 'allowed_chars' are processed
        
        :param item: string to be decoded 
        :return: the decoded string"""    # noqa

        # only allowed chars
        table = {'\\302\\260': '°', '\\303\\240': 'à', '\\303\\242': 'â', '\\303\\244': 'ä', '\\303\\251': 'é',
                 '\\303\\250': 'è', '\\303\\252': 'ê', '\\303\\253': 'ë', '\\303\\257': 'ï', '\\303\\256': 'î',
                 '\\303\\264': 'ô', '\\303\\266': 'ö', '\\303\\271': 'ù', '\\303\\273': 'û', '\\303\\274': 'ü',
                 '\\303\\277': 'ÿ', '\\303\\247': 'ç', '\\303\\261': 'ñ', '\\303\\200': 'À', '\\303\\202': 'Â',
                 '\\303\\204': 'Ä', '\\303\\211': 'É', '\\303\\210': 'È', '\\303\\212': 'Ê', '\\303\\213': 'Ë',
                 '\\303\\217': 'Ï', '\\303\\216': 'Î', '\\303\\224': 'Ô', '\\303\\226': 'Ö', '\\303\\231': 'Ù',
                 '\\303\\233': 'Û', '\\303\\234': 'Ü', '\\305\\270': 'Ÿ', '\\303\\207': 'Ç', '\\303\\221': 'Ñ'}
        new_val = ''
        i = 0
        while i < len(item):
            s = item[i:i+8]
            if s in table:
                new_val += table[s]
                i += 8
            elif item[i:i+2] == '\\\\':
                new_val += '\\'
                i += 2
            else:
                new_val += item[i]
                i += 1
        return new_val

    def remote_ls(self, directory):
        directory = directory or '.'
        LOGGER.debug(f"ssh ls '{directory}' ")
        if self.dryrun:
            return None

        if directory in self.remote_ls_cache.keys():
            return self.remote_ls_cache[directory]

        # TODO: parse also directory / file

        ls_cmd = self.ls_command.format(file=directory)
        try:
            result = self.remote_run(ls_cmd)
        except FileNotFoundError:
            return {}
        res = self.my_decode(result.stdout)
        all_lines = res.split('\n')
        all_files: dict = {}
        for line in all_lines:
            dico = self.get_info_from_format(self.ls_output, line)
            if dico:
                all_files[dico["file"]] = dico

        self.remote_ls_cache[directory] = all_files
        return all_files

    def remote_file_or_dir_exists(self, filepath):
        if self.dryrun:
            return None
        all_files = self.remote_ls(os.path.dirname(filepath))
        if all_files is not None and os.path.basename(filepath) in all_files.keys():
            LOGGER.debug(f"ssh file_exists '{filepath}' : YES")
            return True
        LOGGER.debug(f"ssh file_exists '{filepath}' : NO")
        return False

        # all_files = self.ls(config, os.path.dirname(filepath))   # noqa
        # return os.path.basename(filepath) in all_files.keys()

    def remote_mkdir(self, directory):
        if directory in self.dir_made:
            return
        LOGGER.debug(f"remote mkdir '{directory}'")
        self.remote_run('mkdir -p ' + directory)
        self.dir_made.append(directory)

    def remote_put(self, src, directory):
        LOGGER.debug(f"remote put '{src}' '{directory}'")
        if self.dryrun:
            return
        tmp_file = self.tmp_dir + '/' + src
        if self.connection:
            tmp_path = os.path.dirname(tmp_file)
            self.remote_mkdir(tmp_path)
            self.remote_mkdir(directory)
            sftp = self.connection.sftp()
            sftp.put(src, tmp_file, confirm=True)
            f_a_time = os.path.getatime(src)
            f_m_time = os.path.getmtime(src)
            sftp.utime(tmp_file, (f_a_time, f_m_time))

            # FIXME: is it possible there is a race condition with sftp.put ?

        self.remote_run('mv -vf ' + tmp_file + '  ' + directory)
        if directory in self.remote_ls_cache.keys():
            base = os.path.basename(src)
            dico = {"file": base}  # the only info we have so far, we could add more if necessary
            all_files = self.remote_ls_cache[directory]
            all_files[base] = dico
            self.remote_ls_cache[directory] = all_files  # actually this is useless because the object is shared.

    def remote_get(self, remote_file, local_file):
        LOGGER.debug(f"remote get '{remote_file}' -> '{local_file}'")
        # assuming  directory for local exists
        if self.dryrun:
            return
        if self.connection:
            self.remote_mkdir(self.tmp_dir)
            tmp_file = self.tmp_dir + '/' + os.path.basename(local_file)
            # -p: preserve date -v: verbose -f: force (clobbers)
            self.remote_run(f'cp -pvf {remote_file} {tmp_file}')
            sftp = self.connection.sftp()
            local_dir = os.path.dirname(local_file)
            self.mkdirs(local_dir)
            sftp.get(tmp_file, local_file)
        else:
            self.remote_run(f'mv -vf {remote_file}  {local_file}')

    def remote_rm(self, src: str):
        LOGGER.debug(f"remote rm '{src}'")
        self.remote_run(f'rm -f {src}')

    def remote_rmdir(self, src: str):
        LOGGER.debug(f"remote rmdir '{src}'")
        self.remote_run(f'rm -rf {src}')

    def connect_ssh(self, config: dict):
        """

        :param config: configuration
        :return: Connection_done: bool, uname: str, cause:str, Error: bool
        """
        self.remote_host = config['remote-host']
        self.remote_port = int(config['remote-port'])
        self.remote_user = config['remote-user']
        self.remote_pwd = config['remote-pwd']
        self.ls_output = config['ls-output']
        self.ls_command = config['ls-command']

        if self.connection is not None:
            return True, self.remote_uname, None, False

        remote = (config['enable-remote-copy'] and config['remote-web'] and
                  config['remote-host'] and config['remote-port'])
        if not remote:
            return False, None, 'No remote parameters', False   # not connecting is not error here

        LOGGER.info(f"connect host='{self.remote_host}' port='{self.remote_port}' user='{self.remote_user}'")
        if self.dryrun:
            self.connection = 1  # just to avoid doing connect every time
            return False, None, "Dryrun", False  # not connecting is not error here

        self.connection = fabric.Connection(self.remote_host, self.remote_user, self.remote_port)
        if self.connection is None:
            LOGGER.info("SSH error while Connecting")
            return False, None, "SSH error Connecting", True

        result = self.remote_run('echo $PATH')
        if not result.ok:
            LOGGER.info(f"ssh error {result.exited}")
            return False, None, f"SSH error {result.exited}", True
        res = result.stdout.strip()
        if res == "$PATH":
            return True, "Windows", None, False
        else:
            result = self.remote_run('uname -rsvm')  # noqa

        uname = result.stdout.strip()
        sftp = self.connection.sftp()
        af = sftp.listdir('.')
        if self.tmp_dir in af:
            self.remote_rmdir(self.tmp_dir)
        sftp.mkdir(self.tmp_dir)
        self.remote_uname = uname
        return True, uname, None, False


ACTOR = PwpActor()



# ---------------------------------------------------------------------------------------------------------------
# piwiPre project
# This program and library is licenced under the European Union Public Licence v1.2 (see LICENCE)
# developed by fabien.battini(at)gmail.com
# ---------------------------------------------------------------------------------------------------------------


import sys
import os
import pprint
from piwiPre.pwpErrors import LOGGER
from piwiPre.pwpArgsIni import PwpArgsIni


class PwpParser(PwpArgsIni):
    default_ini_file = "piwiPre.ini"

    def __init__(self, arguments=None, with_config=True, program: str = "piwiPre"):
        super().__init__()
        self.config = None  # the config after parsing HOME,& cwd ini files and cmdline args

        self.add_header("""
Commandline Flags and configuration items
#########################################

This file is the default configuration of piwiPre.

Unless stated otherwise, the  configuration items have a command line argument counterpart,
with the same name, starting with - - .

The default value is given as an argument.

The configuration file uses the yaml syntax,
and uses pyYaml  to read/write the configuration file

- *boolean* values are *true* and *false*
- *None* denotes a value which is not set.
- *string* SHOULD single or double quotes to prevent yaml to interpret values.
- *directory* should be a valid path syntax, written as a string.
- *dictionary* read key : value
""")

        self.add_header("""

Configuration hierarchy
=======================

1. Default values are set for all items.

2. By default, configuration data is read from piwiPre.ini,
   but if the cmdline arguments hold a '--ini-file new-value.ini', then the name of the ini file is changed,
   so that the new ini-file is taken into account.

3. In the user HOME directory, as a special case, '.piwiPre.ini' is read instead.
   This file should be protected against reading from others,
   (chmod 500 in the Linux case).
   It is used to store confidential information:

   - piwigo-host
   - piwigo-port
   - piwigo-user
   - piwigo-pwd

   - remote-host
   - remote-port
   - remote-user
   - remote-pwd
   - remote-incoming 
 
   Other configuration items could also be stored there, but should be preferably stored in cwd or in TRIAGE
  
4. in cwd. Here, one usually  sets up the global configuration without the confidential information
   and without details specific to each TRIAGE directory.
   If this file stores confidential information, it should also be chmod 500.
   Once cwd/piwiPre.ini has been read, new values of the confidential configuration 
   are no more taken into account.

5. On cmdLine. In this case, options start with '--', such as '--version' etc.
 
6. When managing TRIAGE, in TRIAGE subdirectories. 

   These .ini files are read only when processing files in TRIAGE.
   Only directory-specific configuration should be stored here.

   To clarify a difficulty: when managing TRIAGE, the configuration files in ALBUM are *not* read
     
   Typically, one stores there 'names', 'authors', 'dates', 'copyright', 'instructions',
   if some of these should be different for a specific directory.
   Each ini file will be copied in the corresponding ALBUM directories.
   If there was a preexisting .ini file in the ALBUM subdir,
   then it is clobbered by the new one.
   

7. When managing ALBUM, in the directory hierarchy of ALBUM.

   These .ini files are read only when processing files in ALBUM.
   
   To clarify a difficulty: when managing ALBUM, the configuration files in TRIAGE are *not* read

   .ini files in ALBUM are usually a copy of an .ini file in the original TRIAGE directory, but they *can* be
   hand-writen by the piwigo administrator.
    
   .ini files in ALBUM are typically used to maintain directory (nested ALBUM) wide settings.
   
   For instance, some sub-ALBUM may hold pictures without shooting date in the Exif data,
   therefore the naming method is different.


8. The later has precedence over the sooner.
 
9. Therefore, cmdLine items do not modify configuration options found in directories of TRIAGE and ALBUM.
   The only way to reset these are:
   
   - To modify the .ini files in TRIAGE, and then run piwiPre to forward the modifications to ALBUM
   - To edit .ini files in ALBUM 
""")

        self.add_header("""
Some vocabulary
===============

- The **piwigo host** is the server where the piwigo service runs. 
  Usually this is a cloud-based host, or a NAS, or a Linux server.

- The **piwiPre host** is the computer where piwiPre runs. 
  Usually this is a desktop PC, but could be also the same machine than piwigo host.

- **cwd** is the directory where piwiPre is run. If a relative path is used, then it starts from there
""")
        self.add_header("""
cmdLine flags only
==================

The following command line arguments do not have configuration counterpart in the .ini file:
""")
        # -h, --help is implicit

        self.add_item('version',
                      action='store_true',
                      help="Prints piwiPre version number and exits.",
                      location='args',
                      config="""
This flag has no value, it is active only when set""")

        self.add_item('licence',
                      action='store_true',
                      help="prints the LICENCE and exits",
                      location='args',
                      config="""
This flag has no value, it is active only when set""")

        self.add_item('debug',
                      action='store_true',
                      help="Increments the level of verbosity of the logs printed on standard output.",
                      location='args',
                      config="""
This flag has no value, it is active only when set""")

        self.add_item('dump-config',
                      pwp_type=str,
                      action='store',
                      help="Dump the configuration for a given directory and exits.",
                      location='args',
                      config="""
The value of this flag is the name of the directory from which the configuration should be dumped.
  
This path starts from cwd, e.g. TRIAGE/Armor""")

        self.add_item('dryrun',
                      action='store_true',
                      help="Prints what should be done, but does not execute actions.",
                      location='args',
                      config="""this flag has no value, it is active only when set""")

        self.add_item('ini-file',
                      help="Changes the default configuration file to something else.",
                      pwp_type=str,
                      default='piwiPre.ini',
                      action='store',
                      location='args',
                      config="""
The value of this flag is the name of the new configuration file, it will be used in all directories""")

        self.add_item('reset-ini',
                      help="Resets to default the ini file in the current directory and exits.",
                      config="""
This flag has no value, it is active only when set
                      
This is useful to build a new configuration file and edit it.

CAVEAT: previous inifile is clobbered!! """,
                      action='store_true',
                      location='args')

        self.add_item('chdir',
                      help="Changes the default directory where piwiPre is run, is always executed BEFORE --ini-file",
                      pwp_type=str,
                      action='store',
                      location='args',
                      config="""
    The value of this flag is the directory to change to""")

        self.add_header("""
Global actions in ALBUM subdirectories
======================================""")

        self.add_item('verify-albums',
                      help='list of directories in ALBUM to be verified ',
                      action='append',
                      default=[],
                      location='args',
                      pwp_type=list,
                      config="""
   
- Value = a directory in ALBUM to be verified
- Default : [].
- may be used several times

If verify-albums is set, triage and verify-thumbnails are unset

Caveat: sub-directories of the target directory are NOT verified.
""")

        self.add_item('verify-thumbnails',
                      help='list of directories in ALBUM where thumbnails are verified',
                      action='append',
                      default=[],
                      location='args',
                      pwp_type=list,
                      config="""

- Value = a directory in ALBUM to be verified
- Default : [].
- may be used several times

If verify-thumbnails is set, triage is unset (and it means that verify-albums is not set)

Caveat: this is RECURSIVE: sub-directories are managed.
""")

        self.add_item('useless-thumbnails-delete',
                      help='Deletion of useless piwigo thumbnails.',
                      action='store_true',
                      location='args',
                      config="""
When doing verify-albums or verify-thumbnails, 
this flag allows to remove thumbnails that are useless because there is no corresponding picture.

This flag has no value, it is active only when set.
   
- It should be tested first with --dryrun
- It cannot be stored in a configuration file
        """)

        self.add_header("""
.. attention::
    The following flags --test-xxx are used when performing self-testing of piwiPre
    They are not intended to be used under normal circumstances""")

        self.add_item('test-piwigo-synchronization',  # album
                      help="tests piwigo sync through http for the album argument, and exits",
                      default=None,
                      action='store',
                      location='args',
                      config="""
argument = album sub-directory to synchronize
                    """)

        self.add_item('test-ssh', help="tests ssh on remote host and exits",
                      action='store_true',
                      location='args')

        self.add_item('test-sftp', help="tests rcp on remote host (by copying a file in HOME and exits",
                      action='store_true',
                      location='args')

        self.add_header("""
Management of directories
=========================""")

        self.add_item('triage',
                      help='Sets the TRIAGE directory where are stored incoming pictures.',
                      action='store',
                      default='TRIAGE',
                      config="""
- value = 'directory': Sets the TRIAGE directory

  This directory is read-only
  
  This directory can be erased once processing is finished.

- value = None: no TRIAGE directory to process
  
  When verify-albums is used, triage is automatically set to None in order to avoid confusion 
  between the configurations of triage and album
  
""")

        self.add_item('album',
                      help='Sets the root directory for ALBUM where piwigo pictures are stored.',
                      action='store',
                      default='ALBUM',
                      config="""
- value = 'directory' : Sets root directory for ALBUM
 
  a typical value is //NAS/photo 
  
- value =  None, the ALBUM directory is not managed, files are not copied from TRIAGE to ALBUM.""")

        self.add_item('backup',
                      help='Sets the modified directory.',
                      action='store',
                      default='BACKUP',
                      config="""
- value = 'directory' : Sets the BACKUPS directory, where unknown files and modified ALBUM files 
  are saved before  any modification.
  
  This directory can be erased once processing is finished.""")

        self.add_item('web',
                      help='Sets the root directory for piwigo thumbnails.',
                      action='store',
                      default='WEB',
                      config="""
- value = 'directory' : Sets the thumbnails directory.

  a typical value is //NAS/web/piwigo/_data/i/galleries/photo
""")

        self.add_item('remote-web',
                      help='Sets the REMOTE directory for piwigo thumbnails.',
                      action='store',
                      default=None,
                      config="""
- value = 'directory' : Sets the thumbnails directory when accessed through ssh/sftp on the remote host
- if value is None, then piwigo thumbnails are NOT accessed through sftp

a typical value is '/volume1/web/piwigo/_data/i/galleries/photo', appropriate for synology NAS """)

        self.add_header("""
Management of piwigo host and users 
===================================""")

        self.add_item('piwigo-user',
                      help='username on remote server, used for https administration',
                      action='store',
                      default=None,
                      config="""
- Value = 'string' :username on remote server, used by https
- Value = None : anonymous https is assumed""")

        self.add_item('piwigo-pwd',
                      help='Sets the password of the piwigo https admin ',
                      action='store',
                      default=None,
                      location='config')

        self.add_item('piwigo-host',
                      help='sets the hostname of the piwigo server, used by https administration',
                      action='store',
                      default=None,
                      config="""
- Value = 'string' : hostname of the host, used by https
- Value = None : https cannot be used""")

        self.add_item('piwigo-port',
                      help='sets the https port the piwigo server',
                      action='store',
                      pwp_type=int,
                      default=443)
        # --------------------------------------
        # ssh/sftp

        self.add_item('remote-user',
                      help='username on remote server, used for ssh/sftp',
                      action='store',
                      default=None,
                      config="""
- Value = 'string' :username on remote server, used by ssh/sftp
- Value = None : anonymous ssh/sftp is assumed""")

        self.add_item('remote-pwd',
                      help='Sets the password of the piwigo ssh/sftp ',
                      action='store',
                      default=None,
                      location='config')

        self.add_item('remote-host',
                      help='sets the hostname of the piwigo server, used by ssh/sftp',
                      action='store',
                      default=None,
                      config="""
- Value = 'string' : hostname of the host, used by ssh
- Value = None : remote ssh cannot be used""")

        self.add_item('remote-port',
                      help='sets the ssh/sftp port the piwigo server',
                      action='store',
                      pwp_type=int,
                      default=42)

        self.add_item('remote-incoming',
                      help='Path, relative to the remote directory where SFTP launches, where files can be written.',
                      action='store',
                      default=None,
                      config="""
If None, the SFTP root should be writable.

'incoming' is another typical value""")

        self.add_header("""
       
remote host configuration
=========================
Modify these settings only if you know exactly what you are doing.
The default values should be ok with any standard Linux remote host.""")

        self.add_item('ls-command',
                      help='The remote shell command to list files.',
                      action='store',
                      default='ls -sQ1 --full-time {file}',
                      location='config')

        self.add_item('ls-output',
                      help='The output of ls-command.',
                      action='store',
                      default=r'.+\s+{size}\s+{Y}-{m}-{d} {H}:{M}:{S}.\d* {z}\s+"{file}"',
                      location='config',
                      config=r"""
        Where flags are taken from 
        https://docs.python.org/3/library/datetime.html?highlight=datetime#strftime-strptime-behavior ,

           - {dir} is 'd' for a directory
           - {size} is the file size in K Bytes
           - {file} is the file name
           - {Y} is the year, with 4 digits
           - {m} is the month number, with 2 digits
           - {d} is the day number with 2 digits
           - {H} is the hour, with 2 digits
           - {M} the minutes, with 2 digits
           - {S} the seconds, with 2 digits
           - {z} the timezone, expressed as the number of hours and minutes of difference with UTC time, 
             eg. +0100 for CET during winter.
           - {am} is AM or PM

        Alternative for ms-dos, see https://www.windows-commandline.com/get-file-modified-date-time/

           - 'dir {file}'
           - '{Y}/{m}/{d} {H}:{M} {am}'\d*\s+{file}'""")

        self.add_header("""
Management of actions on pictures
=================================

enable-XXX flags have 2 values:

- **false**: no action
- **true** : action is enabled if triage or album mode

By default, all actions are enabled, and this is typically done in the configuration files.


The default values enable a regular processing of ALBUM, provided **verify-albums** is not empty.""")

        self.add_item('enable-rename',
                      help='Enables files renaming',
                      action='store',
                      choices=['true', 'false'],
                      default='true',
                      config="""
In album mode, pictures will **not** be moved from a directory to another, only the filename is changed""")

        self.add_item('enable-rotation',
                      help='Enables picture rotation',
                      action='store',
                      choices=['true', 'false'],
                      default='true',
                      config="""
                      
when ALBUM is moved from photostation to piwigo, since piwigo assumes that pictures are not rotated,
enable-rotation should be used at least once per directory if not done when importing pictures.""")

        self.add_item('enable-metadata',
                      help='Enables the generation of metadata in pictures.',
                      action='store',
                      choices=['true', 'false'],
                      default='true')

        self.add_item('enable-thumbnails',
                      help='Enables generation of Piwigo thumbnails',
                      action='store',
                      choices=['true', 'false'],
                      default='true')

        self.add_item('enable-piwigo-sync',
                      help='Enables the piwigo album synchronization.',
                      action='store',
                      choices=['true', 'false'],
                      default='true')

        self.add_item('enable-remote-copy',
                      help='Enables the copy of piwigo thumbnails with ssh/sftp.',
                      action='store',
                      choices=['true', 'false'],
                      default='false')

        self.add_item('enable-auto-configuration',
                      help='Enables configuration of ALBUM from TRIAGE.',
                      action='store',
                      choices=['true', 'false'],
                      default='true',
                      config="""
Enables the copy of piwiPre.ini files found in TRIAGE directory to the corresponding folder of ALBUM,
so that further processing of ALBUM give the same results.""")

        self.add_header("""
configuration only
==================

The following configuration items are not accessible through command line options
and must be specified in a configuration file.""")

        self.add_item('names',
                      help='The format of renamed pictures. This includes the path starting from ALBUM.',
                      action='store',
                      default='{Y}/{Y}-{m}-{month_name}-{d}-{base}/{Y}-{m}-{d}-{H}h{M}-{S}-{base}.{suffix}',
                      location='config',
                      config=r"""

CAVEAT: The value must be enclosed in single or double quotes !                      
                      
Field values:

- {Y} etc are inherited from the IPTC date of the picture.

- {base} is the name of the TRIAGE folder where the picture was originally found.

- {author} is computed according to the camera name in the IPTC metadata, see **authors**

- {count} is the current count of pictures in the directory,
  so that it is 01 for the first picture, 02 for the 2nd etc.

- {suffix}: file suffix, typically jpg, txt, mp4...

- All numeric fields are printed with 2 digits, excepted year which has 4.

When several  different pictures are supposed to have the same filename,
the last numeric field (here {s}) is incremented until a unique filename is found.


Many users prefer names that include the date,
so that name collisions are avoided when pictures are out in a flat folder.

But different schemes are possible.
For instance, "{Y}/{m}/{d}/{base}-{count}", is also a traditional naming.

all characters that are not in 
"a-zA-Z0-9\-_.&@~!,;+°()àâäéèêëïîôöùûüÿçñÀÂÄÉÈÊËÏÎÔÖÙÛÜŸÇÑ " will be replaced by '_' 
""")  # noqa

        self.add_item('month-name',
                      help='The name for each month, used to compute month_name.',
                      action='store',
                      pwp_type=list,
                      default=['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'],
                      location='config')

        self.add_item('authors',
                      help='A dictionary of mappings between camera model name as found in Exif data, and author name',
                      action='store',
                      pwp_type=dict,
                      default={},
                      location='config',
                      config="""
- example of possible value ::

   Camera1 : author1
   'Camera 2' : 'author 2'
   DEFAULT : 'default value'""")

        self.add_item('copyright',
                      help='A copyright sentence to be written in Exif metadata, with obvious fields.',
                      action='store',
                      default="(C) {author} {Y}",
                      location='config',
                      config="""- The date is taken from the photo metadata, {month} and {day} are also available.'""")

        self.add_item('instructions',
                      help="A sentence to be written in Exif metadata, with {author} coming from the 'authors' section",
                      action='store',
                      default="No copy allowed unless explicitly approved by {author} ",
                      location='config',
                      config="""- adding an email or a phone number may be appropriate.""")

        self.add_item('dates',
                      help='A dictionary of dates corrections',
                      action='store',
                      pwp_type=dict,
                      default='',  # {},
                      location='config',
                      config="""
Date corrections are used only to compute the new name of the picture in the renaming step.
The metadata (notably dates) of the picture is unchanged.

- each toplevel item a dictionary with a unique name
- each date is written as a dictionary with year, month, day, hour, minute, second, some items may be missing

- the dictionary describes each correction with the following fields:

  - 'start', 'end': dates. the correction occurs if the picture date is between these boundaries
  - camera_name: the name of the camera for this correction or 'default' for camera name not in the list
  - 'delta' or 'forced' : a date. 
  
    - If 'delta', the date is added to the picture date. 
    - If  'forced' the picture date is set to this value.

- the specific 'NO-DATE' toplevel item is for pictures without a date.
   - the 'start', 'end', delta dates are not defined
   - this item contains only 'forced' date that will be set to all pictures without a date
- when a date is 'forced', and hour, minute, second are not specified, piwiPre uses the picture creation time.
   
See also the online documentation 
   """)

        self.add_header("""
example ::

    dates:
        USA:                 # this name should be unique within the 'dates'
            start:
                year:  2018
                month:  7
                day: 4
                hour: 20
            end:
                year:  2018
                month:  7
                day: 6
                hour: 23
            D6503:              # camera name
                delta:
                    hour: 9
            TG-320:            # a different camera
                delta:
                    hour: 9
                    minute: 30
        Utah 1:
            start:
                year:  2018
                month:  7
                day: 6
                hour: 23
            end:
                year:  2018
                month:  7
                day: 8
                hour: 23
            TG-320 :
                delta:
                    hours: 8   # CAVEAT: here, hours and not hour ! (and years, etc...)
        NO-DATE:               # CAVEAT: like python, yaml is strict on indentation errors
            forced :
                 year: 2023
                 month: 7
                 day : 24

.. Note:: usually, 'NO-DATE' and  'forced' are not set on a global ALBUM base, 
   but rather in a specific TRIAGE folder where abnormal pictures are known to be stored.


.. Important:: unless enable-auto-configuration false,  
   when a .ini file is stored in a TRIAGE folder, 
   then it  will be copied in the corresponding ALBUM subdirectories, 
   so that further processing of ALBUM give the same results. 
   This is particularly useful for dates management
""")

        self.add_item('piwigo-thumbnails',
                      help="A dictionary of piwigo thumbnails to be built, including formats",
                      action='store',
                      pwp_type=dict,
                      default={
                          "{f}-sq.jpg": {'width': 120, 'height': 120, 'crop': True},
                          "{f}-th.jpg": {'width': 144, 'height': 144, 'crop': False},
                          "{f}-me.jpg": {'width': 792, 'height': 594, 'crop': False},
                          "{f}-cu_e250.jpg": {'width': 250, 'height': 250, 'crop': True},
                      },
                      location='config',
                      config="""
A dictionary if thumbnail specifications,
- {f} is the photo basename
- width = maximum width
- height = maximum height
- crop = the picture will be cropped to a square form factor.

The regular piwigo thumbnails defined in the documentation are as follows ::

    "{f}-sq.jpg" : 120, 120, crop      # SQUARE: mandatory format
    "{f}-th.jpg":  144, 144            # THUMB:  mandatory
    "{f}-me.jpg" : 792, 594            # MEDIUM: mandatory
    "{f}-2s.jpg" : 240, 240            # XXSMALL       # noqa
    "{f}-xs.jpg" : 432, 324            # XSMALL        # noqa
    "{f}-sm.jpg" : 576, 432            # SMALL
    "{f}-la.jpg" : 1008, 756           # LARGE
    "{f}-xl.jpg" : 1224, 918           # XLARGE        # noqa
    "{f}-xx.jpg" : 1656, 1242          # XXLARGE       # noqa
    "{f}-cu_e250.jpg" : 250, 250, crop # CU    : mandatory"""),  # noqa

        # TODO: self.safe_items = ['piwigo-user', 'piwigo-pwd']

        self.config = self.parse_args_and_ini(program, self.default_ini_file, arguments, with_config=with_config)


def build_official_rst(autotest: bool):
    if autotest:
        filename = "tests/results/configuration.rst.autotest"
    else:
        filename = 'source/usage/configuration.rst'
    source = 'piwiPre/pwpParser.py'
    if not autotest and os.path.getmtime(filename) > os.path.getmtime(source):
        LOGGER.msg(f"file '{filename}' is older than source '{source}': patch useless")
        return
    parser = PwpParser(arguments=[], with_config=True, program="autotest")
    LOGGER.msg(f"building official rst '{filename}' from '{source}'")
    parser.build_rst(filename)


def pwp_parser_main(arguments):
    LOGGER.msg('--------------- starting pwp_test_config')
    parser = PwpParser(arguments=arguments, program="parser_autotest", with_config=False)
    config = parser.parse_args_and_ini("test harness", "tests.ini", arguments)
    rst = "../results/test-result.rst"
    ini = "../results/test-result.ini"
    parser.build_rst(rst)
    parser.build_ini_file(ini)
    pprint.pprint(config)
    parser.print_help()
    LOGGER.msg('--------------- end of  pwp_test_config')


if __name__ == "__main__":
    sys.exit(pwp_parser_main(sys.argv[1:]))
